import {Component, EventEmitter, Input, OnInit, Output} from "@angular/core";
import {MatFormField, MatInput} from "@angular/material/input";
import {FormControl, ReactiveFormsModule} from "@angular/forms";
import {MatAutocompleteModule, MatAutocompleteSelectedEvent} from "@angular/material/autocomplete";
import {AsyncPipe} from "@angular/common";
import {catchError, map, Observable, startWith} from "rxjs";
import {LoadingIndicatorComponent} from "@component/common/loading-indicator/loading-indicator.component";
import {ProfessorResponse} from "@api/v1/professorResponse";
import {ProfessorService} from "@api/v1/professor.service";

@Component({
  selector: 'app-professor',
  standalone: true,
  imports: [
    MatAutocompleteModule,
    MatFormField,
    AsyncPipe,
    ReactiveFormsModule,
    MatInput,
    LoadingIndicatorComponent
  ],
  templateUrl: './professor.component.html',
  styleUrl: './professor.component.css',
  providers: [ProfessorService]
})
export class ProfessorComponent implements OnInit {
  protected professorControl = new FormControl();
  protected filteredProfessors!: Observable<ProfessorResponse[]>;

  protected professors: ProfessorResponse[] = [];
  protected professorsLoaded: boolean | null = false;

  @Output() eventResult = new EventEmitter<number>();

  constructor(private api: ProfessorService) {
    this.loadProfessors();
  }

  protected loadProfessors() {
    if (this.professors.length === 0) {
      this.professorsLoaded = false;

      this.api.getProfessors()
        .pipe(catchError(error => {
          this.professorsLoaded = null;
          throw error;
        }))
        .subscribe(data => {
          this.professors = data;
          this.professorsLoaded = true;
        });
    }
  }

  ngOnInit(): void {
    this.filteredProfessors = this.professorControl.valueChanges.pipe(
      startWith(''),
      map(value => this._filterProfessors(value))
    );
  }

  private _filterProfessors(value: string | number): ProfessorResponse[] {
    if (typeof value === 'string') {
      if (value === '') return [];
      const filterValue = value.toLowerCase();
      return this.professors.filter(teacher => teacher.name.toLowerCase().includes(filterValue));
    } else {
      const selectedTeacher = this.professors.find(teacher => teacher.id === value);
      return selectedTeacher ? [selectedTeacher] : [];
    }
  }

  protected onOptionSelected(event: MatAutocompleteSelectedEvent) {
    const selectedOption = this.professors.find(teacher => teacher.id === event.option.value);
    if (selectedOption) {
      this.professorControl.setValue(selectedOption.name);
      this.eventResult.emit(selectedOption.id);
    }
  }
}
